import magic
from pathlib import Path

from deci_common.abstractions.base_model import Schema
from deci_common.data_types.enum.open_vino_checkpoint_types import OpenVinoCheckpointTypes


class OpenVinoCheckpointTypeResolutionError(Exception):
    pass


class OpenVinoCheckpointPointer(Schema):
    checkpoint_path: str
    type: OpenVinoCheckpointTypes = OpenVinoCheckpointTypes.XML

    @staticmethod
    def get_openvino_checkpoint_pointer_from_path(file_path) -> dict:
        """
        Returns a dict of the checkpoint file path and extension
        :param file_path:
        :return: dict
        """
        try:
            file_path_extension = Path(file_path).suffix

            if file_path_extension:
                # REMOVE THE '.' FROM THE FILE EXTENSION
                file_type_extension = file_path_extension.split(".")[-1]
            else:
                # GET THE FILE TYPE FROM THE BINARY MAGIC NUMBER - WORKS ONLY ON NON-UNIX SYSTEMS
                file_type_extension = OpenVinoCheckpointPointer._infer_openvino_file_type_from_binary(file_path)

            pointer = {"checkpoint_path": file_path, "type": file_type_extension}
            return pointer
        except ValueError as e:
            msg = f"The file type is not supported for open vino {file_path}."
            raise Exception(msg) from e

    @staticmethod
    def _infer_openvino_file_type_from_binary(binary_file_path) -> str:
        """
        Helper method to get the file type from the magic number header of the binary (Works only on Unix)
            -- Requires a local installation of libmagic1 on ubuntu
        :param binary_file_path:
        :return:
        """
        try:
            file_type_magic = magic.from_file(binary_file_path, mime=True)

            if ("xml" in file_type_magic) or ("text/plain" in file_type_magic):
                return OpenVinoCheckpointTypes.XML

            elif "zip" in file_type_magic:
                return OpenVinoCheckpointTypes.ZIP

            elif "octet-stream" in file_type_magic:
                return OpenVinoCheckpointTypes.PICKLE

            else:
                raise OpenVinoCheckpointTypeResolutionError(
                    "Could not infer checkpoint file type from binary magic number for: " + str(file_type_magic)
                )

        except OSError as ex:
            msg = "OpenVino file type resolution is supported only on UNIX systems"
            raise OpenVinoCheckpointTypeResolutionError(msg) from ex



import struct
from functools import reduce
from typing import Union, List

import numpy as np

from deci_common.abstractions.abstract_serializer import AbstractTensorSerializer


class NumpySerializer(AbstractTensorSerializer):
    """
    A numpy array serializer/deserializer (like JSON but tiny and fast).
    Effecient to replace JSON and sometimes beats Protobuf in efficiency and memory
    """

    # Warning: Any line of code added here will delay inference runtime.
    # These methods are called 100s of time per second.

    @staticmethod
    def serialize(obj: Union[np.ndarray, List[np.ndarray]]) -> bytes:
        """
        Serializes np.ndarray instances to native bytearrays.
        :param obj: A numpy.ndarray, or a list of np.ndarray instances, to bytes.
        :param data_type: The type of the data in struct library: https://docs.python.org/3/library/struct.html#format-characters
        """
        serialized_bytes: bytes = bytes()
        arrays_to_serialize = obj if isinstance(obj, list) else [obj]
        for array in arrays_to_serialize:
            shape_dims = len(array.shape)
            # Adding the shape dimensions
            serialized_bytes += struct.pack("<I", shape_dims)
            # Adding the shape values
            serialized_bytes += struct.pack("<%uI" % shape_dims, *array.shape)
            # Adding the data type
            serialized_bytes += array.dtype.str.encode()
            # Adding the data length for the tensor.
            # The data length is array.dtype.itemsize times each of the elements in the array shape.
            tensor_size_in_bytes = reduce(lambda a, b: a * b, [array.dtype.itemsize] + list(array.shape))
            serialized_bytes += struct.pack("<I", tensor_size_in_bytes)
            # Adding array raw bytes
            serialized_bytes += array.data
        return serialized_bytes

    @staticmethod
    def deserialize(obj: bytes) -> Union[np.ndarray, List[np.ndarray]]:
        """
        Deserializes native buffer instance to np.ndarray instance(s).
        :param obj:
        :param obj: A numpy.ndarray, or a list of np.ndarray instances, to bytes.
        """
        obj = memoryview(obj)
        input_buffer_length = len(obj)
        deserialized_tensors = []
        deserialize_index = 0
        deserialized_all = False
        while not deserialized_all:
            # Deserializing the shape
            shape_dims: int = struct.unpack("<I", obj[deserialize_index : deserialize_index + 4])[0]
            deserialize_index += 4
            shape: tuple = struct.unpack(
                "<%uI" % shape_dims, obj[deserialize_index : deserialize_index + 4 * shape_dims]
            )
            deserialize_index += 4 * shape_dims

            # Deserializing the np.dtype
            dtype_length = 3
            dtype_bytes = obj[deserialize_index : deserialize_index + dtype_length].tobytes()
            dtype = np.dtype(dtype_bytes)
            deserialize_index += dtype_length

            # Deserializing the data length
            tensor_length_field_size = 4
            next_index = deserialize_index + tensor_length_field_size
            tensor_size_in_bytes = struct.unpack("<I", obj[deserialize_index:next_index].tobytes())[0]
            deserialize_index += tensor_length_field_size

            # Deserializing the data
            data = obj[deserialize_index : deserialize_index + tensor_size_in_bytes]
            tensor = np.frombuffer(data, dtype=np.dtype(dtype)).reshape(shape)
            deserialize_index += tensor_size_in_bytes
            deserialized_tensors.append(tensor)

            # Check if we went through the entire buffer, or are there remaining objects:
            if deserialize_index >= input_buffer_length:
                deserialized_all = True
        if len(deserialized_tensors) == 1:
            # Returning a single np.ndarray, if there is only 1 element in the list;
            # Models should receive
            return deserialized_tensors[0]
        return deserialized_tensors



from enum import Enum

# TODO: rename HardwareType to HardwareName and inferenceHardware to HardwareType
from deci_common.data_types.enum.models_enums import BatchSize, BatchSizeEdge
from deci_common.helpers import get_enum_values


class HardwareType(str, Enum):
    """
    The type of the hardware to run on (CPU/GPU Names)
    """

    K80 = "K80"
    V100 = "V100"
    T4 = "T4"
    CPU = "CPU"
    EPYC = "EPYC"
    XAVIER = "Jetson Xavier"
    NANO = "Jetson Nano"
    XAVIER_AGX = "Jetson Xavier AGX"
    CASCADE_LAKE = "Cascade Lake"
    SKYLAKE = "Skylake"
    Broadwell = "Broadwell"
    NUC_TIGER_LAKE = "Intel NUC Tiger Lake"


class HardwareMachineModel(str, Enum):
    K80 = "p2.xlarge"
    V100 = "p3.2xlarge"
    T4 = "g4dn.2xlarge"
    CPU = "c5.4xlarge"
    EPYC = "c5a.2xlarge"
    NANO = "Jetson Nano"
    XAVIER = "Jetson Xavier"
    XAVIER_AGX = "Jetson Xavier AGX"
    CASCADE_LAKE = "c5.4xlarge"
    SKYLAKE = "c5n.4xlarge"
    Broadwell = "m4.4xlarge"
    NUC_TIGER_LAKE = "Intel NUC Tiger Lake"


class InferenceHardware(str, Enum):
    """
    Hardware that can be used for deep learning inference.
    """

    CPU = "cpu"
    GPU = "gpu"


class MapHardwareTypeToFamily(str, Enum):
    # GPUs
    V100 = InferenceHardware.GPU.value
    K80 = InferenceHardware.GPU.value
    T4 = InferenceHardware.GPU.value
    NANO = InferenceHardware.GPU.value
    XAVIER = InferenceHardware.GPU.value
    XAVIER_AGX = InferenceHardware.GPU.value

    # CPUs
    CPU = InferenceHardware.CPU.value
    EPYC = InferenceHardware.CPU.value
    CASCADE_LAKE = InferenceHardware.CPU.value
    SKYLAKE = InferenceHardware.CPU.value
    Broadwell = InferenceHardware.CPU.value
    NUC_TIGER_LAKE = InferenceHardware.CPU.value


class InferyVersion(str, Enum):
    CPU = "cpu"
    GPU = "gpu"
    JETSON = "jetson"


class MapHardwareTypeToInferyVersion(str, Enum):
    # GPUs
    V100 = InferyVersion.GPU.value
    K80 = InferyVersion.GPU.value
    T4 = InferyVersion.GPU.value

    # Jetsons
    NANO = InferyVersion.JETSON.value
    XAVIER = InferyVersion.JETSON.value
    XAVIER_AGX = InferyVersion.JETSON.value

    # CPUs
    CPU = InferyVersion.CPU.value
    EPYC = InferyVersion.CPU.value
    CASCADE_LAKE = InferyVersion.CPU.value
    SKYLAKE = InferyVersion.CPU.value
    Broadwell = InferyVersion.CPU.value
    NUC_TIGER_LAKE = InferyVersion.CPU.value


class HardwareEnvironment(str, Enum):
    GCP = "gcp"
    AWS = "aws"
    Azure = "azure"
    PREMISE = "on premise"


class HardwareVendor(str, Enum):
    INTEL = "intel"
    NVIDIA = "nvidia"
    AMD = "amd"


class HardwareTaint(str, Enum):
    # GPUs
    V100 = "nvidia.com/v100gpu"
    K80 = "nvidia.com/k80gpu"
    T4 = "nvidia.com/t4gpu"
    NANO = "nvidia.com/jetson-nano"
    XAVIER = "nvidia.com/jetson-xavier"
    XAVIER_AGX = "nvidia.com/jetson-xavier-agx"
    # CPUs
    EPYC = "amd.com/epyc"
    CASCADE_LAKE = "intel.com/cascade-lake"
    SKYLAKE = "intel.com/skylake"
    Broadwell = "intel.com/broadwell"
    NUC_TIGER_LAKE = "intel.com/nuc-tiger-lake"


class HardwareLabel(str, Enum):
    # GPUs
    V100 = "nvidia-tesla-v100"
    K80 = "nvidia-tesla-k80"
    T4 = "nvidia-tesla-t4"
    M60 = "nvidia-m60"
    NANO = "nvidia-jetson-nano"
    XAVIER = "nvidia-jetson-xavier"
    XAVIER_AGX = "nvidia-jetson-xavier-agx"

    # CPUs
    EPYC = "amd-epyc"
    CASCADE_LAKE = "intel-cascade-lake"
    SKYLAKE = "intel-skylake"
    Broadwell = "intel-broadwell"
    NUC_TIGER_LAKE = "intel-nuc-tiger-lake"


class MapHardwareTypeToVendor(str, Enum):
    V100 = HardwareVendor.NVIDIA.value
    K80 = HardwareVendor.NVIDIA.value
    T4 = HardwareVendor.NVIDIA.value
    EPYC = HardwareVendor.AMD.value
    NANO = HardwareVendor.NVIDIA.value
    XAVIER = HardwareVendor.NVIDIA.value
    XAVIER_AGX = HardwareVendor.NVIDIA.value
    CASCADE_LAKE = HardwareVendor.INTEL.value
    SKYLAKE = HardwareVendor.INTEL.value
    Broadwell = HardwareVendor.INTEL.value
    NUC_TIGER_LAKE = HardwareVendor.INTEL.value


class MapHardwareTypeToEnvironment(str, Enum):
    # AWS
    V100 = HardwareEnvironment.AWS.value
    K80 = HardwareEnvironment.AWS.value
    T4 = HardwareEnvironment.AWS.value
    EPYC = HardwareEnvironment.AWS.value
    CASCADE_LAKE = HardwareEnvironment.AWS.value
    SKYLAKE = HardwareEnvironment.AWS.value
    Broadwell = HardwareEnvironment.AWS.value

    # PREMISE
    NANO = HardwareEnvironment.PREMISE.value
    XAVIER = HardwareEnvironment.PREMISE.value
    XAVIER_AGX = HardwareEnvironment.PREMISE.value
    NUC_TIGER_LAKE = HardwareEnvironment.PREMISE.value


class MapHardwareTypeToDefaultBatchSizeList(list, Enum):
    V100 = get_enum_values(BatchSize)
    K80 = get_enum_values(BatchSize)
    T4 = get_enum_values(BatchSize)
    EPYC = get_enum_values(BatchSize)
    NANO = get_enum_values(BatchSizeEdge)
    XAVIER = get_enum_values(BatchSizeEdge)
    XAVIER_AGX = get_enum_values(BatchSizeEdge)
    CASCADE_LAKE = get_enum_values(BatchSize)
    SKYLAKE = get_enum_values(BatchSize)
    Broadwell = get_enum_values(BatchSize)
    NUC_TIGER_LAKE = get_enum_values(BatchSize)


class HardwareImageRepository(str, Enum):
    INTEL = "intel"
    JETSON = "jetson"


class MapHardwareTypeToImageRepository(str, Enum):
    NANO = HardwareImageRepository.JETSON.value
    XAVIER = HardwareImageRepository.JETSON.value
    XAVIER_AGX = HardwareImageRepository.JETSON.value
    NUC_TIGER_LAKE = HardwareImageRepository.INTEL.value


class HardwareImageDistribution(str, Enum):
    J46 = "j46"


class MapHardwareTypeToImageDistribution(str, Enum):
    NANO = HardwareImageDistribution.J46.value
    XAVIER = HardwareImageDistribution.J46.value
    XAVIER_AGX = HardwareImageDistribution.J46.value



from enum import Enum

from deci_common.data_types.enum.hardware_enums import InferenceHardware


class FrameworkFileExtension(Enum):
    TENSORFLOW1 = [".pb"]
    TENSORFLOW2 = [".zip"]
    PYTORCH = [".pth", ".pt"]
    ONNX = [".onnx"]
    TENSORRT = [".pkl"]
    OPENVINO = [".pkl"]
    TORCHSCRIPT = [".pth", ".pt"]
    TVM = None
    KERAS = [".h5"]
    TFLITE = [".tflite"]
    COREML = [".mlmodel"]


class FrameworkType(str, Enum):
    """
    A general deep learning framework, without a version.
    """

    TENSORFLOW1 = "tf1"
    TENSORFLOW2 = "tf2"
    PYTORCH = "pytorch"
    ONNX = "onnx"
    TORCHSCRIPT = "torchscript"
    TENSORRT = "trt"
    TVM = "tvm"
    OPENVINO = "openvino"
    KERAS = "keras"
    TFLITE = "tflite"
    COREML = "coreml"

    @staticmethod
    def from_string(framework: str) -> Enum:
        framework = framework.lower()
        return FrameworkType(framework)


def get_hardware_families_by_framework_type(framework: FrameworkType):
    if framework == FrameworkType.OPENVINO:
        return [InferenceHardware.CPU]
    elif framework == FrameworkType.TENSORRT:
        return [InferenceHardware.GPU]
    else:
        return [InferenceHardware.GPU, InferenceHardware.CPU]



from typing import Type, Dict, List, Any, TypeVar, Generic
from uuid import UUID

from pymongo import ReturnDocument

from deci_common.abstractions.base_model import DBSchema
from deci_common.data_connection.mongo_db_connector import MongoDBConfiguration, MongoDBConnector
from deci_common.abstractions.abstract_logger import ILogger


T = TypeVar("T", bound=DBSchema)


class DBInterface(Generic[T], ILogger):
    def __init__(self, element_type: Type[T], mongo_db_configuration: Type[MongoDBConfiguration]):
        super().__init__()
        self._element_type = element_type
        self._db_connector = MongoDBConnector(mongo_db_configuration)

    def insert(self, element: T) -> T:
        """
        Insert an element to the DB and return it
        @param element: The element to insert
        @return: T
        """
        with self._db_connector as db:
            db._insert_document(element.dict())
        return element

    def find(self, search_filter: Dict[str, Any] = {}) -> List[T]:
        """
        Find all elements from DB with a specified search filter
        @param search_filter: Filter for the search
        @return: List[T]
        """
        with self._db_connector as db:
            elements = db._get_documents(search_filter=search_filter)

        return [self._element_type(**element) for element in elements]

    def update(self, element: T) -> T:
        """
        Updates an element in the DB using its id
        @param element: The element to update to (id is used to find the element to update)
        @return: T
        """
        with self._db_connector as db:
            updated_element = db._update_and_return_document(
                search_filter={"id": element.id}, updated_document=element.dict(), return_document=ReturnDocument.AFTER
            )
        return self._element_type(**updated_element)

    def delete(self, id: UUID, force: bool = False) -> bool:
        """
        Deletes an element from the DB
        @param id: The ID of the element to delete
        @param force: if True, permanently delete the element from the DB, otherwise - just mark it as deleted
        @return: True on success, False otherwise
        """
        with self._db_connector as db:
            if force:
                delete_result = db._delete_document(search_filter={"id": id})
                return delete_result.deleted_count == 1
            updated_element = db._update_and_return_document(
                search_filter={"id": id}, updated_document={"deleted": True}, return_document=ReturnDocument.AFTER
            )
            return updated_element["deleted"]



import logging
import logging.config

from deci_common.auto_logging import AutoLoggerConfig


def get_logger(logger_name: str) -> logging.Logger:
    config_dict = AutoLoggerConfig.generate_config_for_module_name(logger_name)
    logging.config.dictConfig(config_dict)
    logger: logging.Logger = logging.getLogger(logger_name)
    return logger


class ILogger:
    """
    Provides logging capabilities to the derived class.
    """

    def __init__(self, logger_name: str = None):
        logger_name = logger_name if logger_name else str(self.__module__)
        self._logger: logging.Logger = get_logger(logger_name)



from abc import abstractmethod
from pymongo.collection import ReturnDocument

from deci_common.abstractions.abstract_logger import ILogger


class AbstractDocumentDBConnector(ILogger):
    """
    Provides operations on document databases.
    """

    @abstractmethod
    def _insert_document(self, document_json: dict):
        raise NotImplementedError()

    @abstractmethod
    def _get_document(self, search_filter: dict) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def _update_document(self, search_filter: dict, updated_document: dict, upsert: bool = False):
        raise NotImplementedError()

    @abstractmethod
    def _update_and_return_document(
        self,
        search_filter: dict,
        updated_document: dict,
        upsert: bool = False,
        return_document: ReturnDocument = ReturnDocument.BEFORE,
    ):
        raise NotImplementedError()

    @abstractmethod
    def _delete_document(self, search_filter: dict):
        raise NotImplementedError()



import datetime
from abc import ABCMeta
from typing import Type, Dict, Any

from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.collection import ReturnDocument

from deci_common.abstractions.abstract_document_db_connector import AbstractDocumentDBConnector


class MongoDBConfiguration:
    """
    The base of the mongo db configurations - Classes may derive from it and change only the 'COLLECTION'.
    """

    __metaclass__ = ABCMeta

    DATABASE_NAME = None
    HOST = None
    PORT = None
    USERNAME = None
    PASSWORD = None
    AUTH_DB = None  # The database on which will preform the authentication
    COLLECTION = None
    PEM_FILE_URL = None  # The pem file for TLS connection stored inside AWS SecretsManager


class MongoDBConnector(AbstractDocumentDBConnector):
    """
    An abstract MongoDBConnector that operates on a collection (SQL Table equivalent).
    """

    def __init__(self, mongo_db_configuration: Type[MongoDBConfiguration]):
        assert issubclass(mongo_db_configuration, MongoDBConfiguration)
        self._db_config: Type[MongoDBConfiguration] = mongo_db_configuration
        self._client: MongoClient = None
        super().__init__()

    @property
    def client(self) -> MongoClient:
        if not self._client:
            raise RuntimeError('The client is not initialized. You must use the client withing a "with" clause')
        return self._client

    def __create_client(self) -> MongoClient:
        """
        Creates a new pymongo.MongoClient for DB operations.
        """
        # NOTE: This is done here to avoid circular dependency when loading this module together with the
        # FilesDataInterface module

        from deci_common.data_interfaces.files_data_interface import FilesDataInterface

        client = MongoClient(
            host=self._db_config.HOST,
            port=self._db_config.PORT,
            username=self._db_config.USERNAME,
            password=self._db_config.PASSWORD,
            uuidRepresentation="pythonLegacy",
            retryWrites=False,
            connect=True,
            tls=True,
            tlsCAFile=FilesDataInterface.download_temporary_file(self._db_config.PEM_FILE_URL),
        )
        self._logger.debug("Authenticated to db successfully.")

        return client

    def _get_collection(self, collection_name: str = None) -> Collection:
        """
        Returns a specific collection (NoSQL Equivalent to Table in SQL, by the collection's name. :param
        collection_name: The name of the collection (table) to fetch. Default it the collection from the
        MongoDbConfiguration
        """
        if not collection_name:
            collection_name = self._db_config.COLLECTION
        db = self.client[self._db_config.DATABASE_NAME]
        return db[collection_name]

    def _delete_document(self, search_filter: dict):
        """
        Deletes a document from the collection.
        """
        collection = self._get_collection()
        return collection.delete_one(filter=search_filter)

    def _insert_document(self, document_json: dict):
        now = datetime.datetime.utcnow()
        document_json.update({"creation_time": now, "update_time": now})
        collection = self._get_collection()
        return collection.insert_one(document=document_json)

    def _get_document(self, search_filter: dict) -> dict:
        collection = self._get_collection()
        doc = collection.find_one(filter=search_filter)
        return doc

    def _get_documents(self, search_filter: Dict[str, Any], project_fields: Dict[str, Any] = {}) -> Cursor:
        collection = self._get_collection()
        amount = collection.count_documents(search_filter)
        if not amount:
            return []
        documents_cursor = collection.find(filter=search_filter, projection=project_fields)
        if documents_cursor:
            return documents_cursor
        return []

    def _update_document(self, search_filter: dict, updated_document: dict, upsert: bool = False):
        updated_document.update({"update_time": datetime.datetime.utcnow()})
        collection = self._get_collection()
        return collection.update_one(filter=search_filter, update={"$set": updated_document}, upsert=upsert)

    def _update_and_return_document(
        self,
        search_filter: dict,
        updated_document: dict,
        upsert: bool = False,
        return_document: ReturnDocument = ReturnDocument.BEFORE,
    ):
        updated_document.update({"update_time": datetime.datetime.utcnow()})
        collection = self._get_collection()
        return collection.find_one_and_update(
            filter=search_filter, update={"$set": updated_document}, upsert=upsert, return_document=return_document
        )

    def _remove_from_list_and_return_document(self, search_filter: dict, remove_query: dict):
        collection = self._get_collection()
        return collection.find_one_and_update(
            filter=search_filter, update={"$pull": remove_query, "$set": {"update_time": datetime.datetime.utcnow()}}
        )

    def _add_to_list_and_return_document(self, search_filter: dict, add_query: dict):
        collection = self._get_collection()
        return collection.find_one_and_update(
            filter=search_filter,
            update={"$addToSet": add_query, "$set": {"update_time": datetime.datetime.utcnow()}},
            upsert=True,
        )

    def _increment_a_field(self, search_filter: dict, increment_query: dict):
        collection = self._get_collection()
        collection.update_one(
            filter=search_filter,
            update={"$inc": increment_query, "$set": {"update_time": datetime.datetime.utcnow()}},
            upsert=True,
        )

    def ping(self):
        db = self.client.get_database(self._db_config.DATABASE_NAME)
        return db.command("ping")

    def __enter__(self):
        if not self._client:
            self._client = self.__create_client()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
        # The following lines will close the connection after __exit__, or after each 'with' statement.
        # Disabling this behaviour and keeping the client alive all the time (self._client initialized once).
        # self._client.close()
        # self._client = None



from enum import Enum


class DeepLearningTask(str, Enum):
    CLASSIFICATION = "classification"
    SEMANTIC_SEGMENTATION = "semantic_segmentation"
    OBJECT_DETECTION = "object_detection"
    NLP = "nlp"
    OTHER = "other"
